
import unittest

from uvm.base.uvm_cmdline_processor import UVMCmdlineProcessor
from uvm.base.uvm_debug import UVMDebug

UVMDebug.DEBUG = False

class TestUVMCmdLineProc(unittest.TestCase):
    """ Unit tests for UVMCmdLineProcessor"""

    def test_get_arg(self):
        UVMCmdlineProcessor.m_test_mode = False
        #UVMCmdlineProcessor.m_test_plusargs = {"ABC": "123"}
        argv = ["+ABC=123"]
        clp = UVMCmdlineProcessor()
        clp.extract_args(argv)
        arr = []
        count = clp.get_arg_values("+MY_TEST", arr)
        self.assertEqual(count, 0)
        count = clp.get_arg_values("+ABC=", arr)
        self.assertEqual(count, 1)
        self.assertEqual(arr[0], "123")
        UVMCmdlineProcessor.m_test_mode = False

    def test_get_arg_matches(self):
        UVMCmdlineProcessor.m_test_mode = False
        plusargs = {"ABC": "123", "WW": "XX", "TRACE": "123"}
        argv = []
        for i, (key,val) in enumerate(plusargs.items()):
            argv.append("+" + key + "=" + str(val))
        #UVMDebug.DEBUG = True
        clp = UVMCmdlineProcessor()
        clp.extract_args(argv)
        args = []
        num = clp.get_arg_matches("XYZ", args)
        self.assertEqual(num, 0)
        num = clp.get_arg_matches("+TRACE=", args)
        self.assertEqual(num, 1)
        num = clp.get_arg_matches("+TRACE", args)
        self.assertEqual(num, 1)
        UVMCmdlineProcessor.m_test_mode = False

    def test_multiple_args(self):
        UVMCmdlineProcessor.m_test_mode = False
        UVMCmdlineProcessor.m_test_plusargs = {}
        argv = ["+ABC=123", "+ABC=567"]
        UVMCmdlineProcessor.m_test_argv = argv
        clp = UVMCmdlineProcessor()
        clp.extract_args(argv)
        args = []
        num = clp.get_arg_matches("+ABC=", args)
        self.assertEqual(num, 2)
        print('2 matches are' + str(args))

    def test_regex_in_arg_matches(self):
        UVMCmdlineProcessor.m_test_mode = False
        plusargs = {"ABC": "123", "WW": "XX", "TRACE": "123", "COMPLEX_MATCH": "555"}
        argv = []
        for i, (key,val) in enumerate(plusargs.items()):
            argv.append("+" + key + "=" + str(val))
        clp = UVMCmdlineProcessor()
        clp.extract_args(argv)
        args = []
        num = clp.get_arg_matches("/^XYZ$/", args)
        self.assertEqual(num, 0)

        args = []
        num = clp.get_arg_matches("/^\\+COMPLEX_MATCH=$/", args)
        self.assertEqual(num, 1)
        UVMCmdlineProcessor.m_test_mode = False




if __name__ == '__main__':
    unittest.main()



#
#------------------------------------------------------------------------------
#   Copyright 2007-2010 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

import random

from .uvm_object import UVMObject
from .uvm_object_globals import (UVM_WARNING, UVM_INFO, UVM_ERROR, UVM_FATAL, UVM_LOW, UVM_NONE,
        UVM_MEDIUM)
from .uvm_report_message import UVMReportMessage


def get_verbosity(severity):
    if severity == UVM_ERROR:
        return UVM_LOW
    elif severity == UVM_FATAL:
        return UVM_NONE
    else:
        return UVM_MEDIUM





class UVMReportObject(UVMObject):
    """
    CLASS: UVMReportObject

    The uvm_report_object provides an interface to the UVM reporting facility.
    Through this interface, components issue the various messages that occur
    during simulation. Users can configure what actions are taken and what
    file(s) are output for individual messages from a particular component
    or for all messages from all components in the environment. Defaults are
    applied where there is no explicit configuration.

    Most methods in uvm_report_object are delegated to an internal instance of a
    <uvm_report_handler>, which stores the reporting configuration and determines
    whether an issued message should be displayed based on that configuration.
    Then, to display a message, the report handler delegates the actual
    formatting and production of messages to a central <uvm_report_server>.

    A report consists of an id string, severity, verbosity level, and the textual
    message itself. They may optionally include the filename and line number from
    which the message came. If the verbosity level of a report is greater than the
    configured maximum verbosity level of its report object, it is ignored.
    If a report passes the verbosity filter in effect, the report's action is
    determined. If the action includes output to a file, the configured file
    descriptor(s) are determined.

    Actions - can be set for (in increasing priority) severity, id, and
    (severity,id) pair. They include output to the screen <UVM_DISPLAY>,
    whether the message counters should be incremented <UVM_COUNT>, and
    whether a $finish should occur <UVM_EXIT>.

    Default Actions - The following provides the default actions assigned to
    each severity. These can be overridden by any of the ~set_*_action~
    methods::
        UVM_INFO -       UVM_DISPLAY
        UVM_WARNING -    UVM_DISPLAY
        UVM_ERROR -      UVM_DISPLAY | UVM_COUNT
        UVM_FATAL -      UVM_DISPLAY | UVM_EXIT

    File descriptors - These can be set by (in increasing priority) default,
    severity level, an id, or (severity,id) pair.  File descriptors are
    standard SystemVerilog file descriptors; they may refer to more than one file.
    It is the user's responsibility to open and close them.

    Default file handle - The default file handle is 0, which means that reports
    are not sent to a file even if a UVM_LOG attribute is set in the action
    associated with the report. This can be overridden by any of the ~set_*_file~
    methods.
    """

    def __init__(self, name=""):
        UVMObject.__init__(self, name)
        from .uvm_report_handler import UVMReportHandler
        self.m_rh = UVMReportHandler(name)

    #----------------------------------------------------------------------------
    # Group: Reporting
    #----------------------------------------------------------------------------

    def uvm_get_report_object(self):
        """
        Returns the nearest uvm_report_object when called.  From inside a
        `UVMComponent`, the method simply returns `self`.

        See also the global version of `uvm_get_report_object`.
        Returns:
        """
        return self

    def uvm_report_enabled(self, verbosity, severity=UVM_INFO, id=""):
        """
        Returns 1 if the configured verbosity for this severity/id is greater than
        or equal to `verbosity` else returns 0.

        See also `get_report_verbosity_level` and the global version of
        `uvm_report_enabled`.
        Args:
            verbosity:
            severity:
            id:
        Returns:
            bool: True if given report enabled, False otherwise.
        """
        if self.get_report_verbosity_level(severity, id) >= verbosity:
            return True
        return False

    def uvm_report(self, severity, id, message,
            verbosity=-1,
            filename="",
            line=0,
            context_name="",
            report_enabled_checked=False):

        rng_state = random.getstate()  # tpoikela: Added to ensure logging does
        # not affect Random Stability
        if verbosity == -1:
            verbosity = get_verbosity(severity)
        l_report_message = None
        if report_enabled_checked is False:
            if not self.uvm_report_enabled(verbosity, severity, id):
                return
        l_report_message = UVMReportMessage.new_report_message()
        l_report_message.set_report_message(severity, id, message,
                                            verbosity, filename, line, context_name)
        self.uvm_process_report_message(l_report_message)
        random.setstate(rng_state)

    def uvm_report_info(self, id, message, verbosity=UVM_MEDIUM, filename="",
            line=0, context_name="", report_enabled_checked=False):
        """
        Function: uvm_report_info

        Args:
            id (str): Message ID/tag.
            message (str): Message string.
            verbosity (int): Verbosity of the message.
            filename (str): Filename of the caller.
            line (int): Line from where this is called.
            context_name (str): Name of the calling context.
            report_enabled_checked (bool): Checks report enabled.
        """
        self.uvm_report(UVM_INFO, id, message, verbosity,
                  filename, line, context_name, report_enabled_checked)

    def uvm_report_warning(self, id, message, verbosity=UVM_MEDIUM, filename="",
            line=0, context_name="", report_enabled_checked=False):
        """
        Function: uvm_report_warning
        Args:
            id:
            message:
            verbosity:
            filename:
            line:
            context_name:
            report_enabled_checked:
        """
        self.uvm_report(UVM_WARNING, id, message, verbosity,
                filename, line, context_name, report_enabled_checked)

    def uvm_report_error(self, id, message, verbosity = UVM_LOW, filename = "",
            line = 0, context_name = "", report_enabled_checked = False):
        """
        Function: uvm_report_error
        Args:
            id:
            message:
            verbosity:
            filename:
            line:
            context_name:
            report_enabled_checked:
        """
        self.uvm_report (UVM_ERROR, id, message, verbosity,
                filename, line, context_name, report_enabled_checked)

    def uvm_report_fatal(self, id, message, verbosity = UVM_NONE, filename = "",
            line = 0, context_name = "", report_enabled_checked = False):
        """
        Function: uvm_report_fatal
        Args:
            id:
            message:
            verbosity:
            filename:
            line:
            context_name:
            report_enabled_checked:
        """
        self.uvm_report(UVM_FATAL, id, message, verbosity,
                filename, line, context_name, report_enabled_checked)

    def uvm_process_report_message(self, report_message):
        """
        Function: uvm_process_report_message

        This method takes a preformed uvm_report_message, populates it with
        the report object and passes it to the report handler for processing.
        It is expected to be checked for verbosity and populated.
        Args:
            report_message:
        """
        report_message.set_report_object(self)
        self.m_rh.process_report_message(report_message)

    #----------------------------------------------------------------------------
    # Group: Verbosity Configuration
    #----------------------------------------------------------------------------

    def get_report_verbosity_level(self, severity=UVM_INFO, id=""):
        """
        Gets the verbosity level in effect for this object. Reports issued
        with verbosity greater than this will be filtered out. The severity
        and tag arguments check if the verbosity level has been modified for
        specific severity/tag combinations.

        Args:
            severity:
            id (str):
        Returns:
        """
        return self.m_rh.get_verbosity_level(severity, id)

    def get_report_max_verbosity_level(self):
        """
        Function: get_report_max_verbosity_level

        Gets the maximum verbosity level in effect for this report object.
        Any report from this component whose verbosity exceeds this maximum will
        be ignored.

        Returns:
            int: Maximum verbosity level.
        """
        return self.m_rh.m_max_verbosity_level

    def set_report_verbosity_level(self, verbosity_level):
        """
        Function: set_report_verbosity_level

        This method sets the maximum verbosity level for reports for this component.
        Any report from this component whose verbosity exceeds this maximum will
        be ignored.

        Args:
            verbosity_level:
        """
        self.m_rh.set_verbosity_level(verbosity_level)

    def set_report_id_verbosity(self, id, verbosity):
        """
        Function: set_report_id_verbosity

        Args:
            id:
            verbosity:
        """
        self.m_rh.set_id_verbosity(id, verbosity)

    # Function: set_report_severity_id_verbosity
    #
    # These methods associate the specified verbosity threshold with reports of the
    # given ~severity~, ~id~, or ~severity-id~ pair. This threshold is compared with
    # the verbosity originally assigned to the report to decide whether it gets
    # processed.    A verbosity threshold associated with a particular ~severity-id~
    # pair takes precedence over a verbosity threshold associated with ~id~, which
    # takes precedence over a verbosity threshold associated with a ~severity~.
    #
    # The ~verbosity~ argument can be any integer, but is most commonly a
    # predefined <uvm_verbosity> value, <UVM_NONE>, <UVM_LOW>, <UVM_MEDIUM>,
    # <UVM_HIGH>, <UVM_FULL>.

    def set_report_severity_id_verbosity(self, severity, id, verbosity):
        self.m_rh.set_severity_id_verbosity(severity, id, verbosity)

    #----------------------------------------------------------------------------
    # Group: Action Configuration
    #----------------------------------------------------------------------------

    # Function: get_report_action
    #
    # Gets the action associated with reports having the given ~severity~
    # and ~id~.

    def get_report_action(self, severity, id):
        return self.m_rh.get_action(severity,id)

    def set_report_severity_action(self, severity, action):
        """
        Function: set_report_severity_action

        Args:
            severity:
            action:
        """
        self.m_rh.set_severity_action(severity, action)

    def set_report_id_action(self, id, action):
        """
        Function: set_report_id_action

        Args:
            id:
            action:
        """
        self.m_rh.set_id_action(id, action)

    def set_report_severity_id_action(self, severity, id, action):
        """
        Function: set_report_severity_id_action

        These methods associate the specified action or actions with reports of the
        given `severity`, `id`, or ~severity-id~ pair. An action associated with a
        particular ~severity-id~ pair takes precedence over an action associated with
        `id`, which takes precedence over an action associated with a `severity`.

        The `action` argument can take the value `UVM_NO_ACTION`, or it can be a
        bitwise OR of any combination of `UVM_DISPLAY`, `UVM_LOG`, `UVM_COUNT`,
        `UVM_STOP`, `UVM_EXIT`, and `UVM_CALL_HOOK`.
        Args:
            severity:
            id:
            action:
        """
        self.m_rh.set_severity_id_action(severity, id, action)

    #----------------------------------------------------------------------------
    # Group: File Configuration
    #----------------------------------------------------------------------------

    def get_report_file_handle(self, severity, id):
        """
        Function: get_report_file_handle

        Gets the file descriptor associated with reports having the given
        `severity` and `id`.
        Args:
            severity:
            id:
        Returns:
        """
        return self.m_rh.get_file_handle(severity,id)

    def set_report_default_file(self, file):
        """
        Function: set_report_default_file
        Args:
            file:
        """
        self.m_rh.set_default_file(file)

    def set_report_id_file(self, id, file):
        """
        Function: set_report_id_file
        Args:
            id:
            file:
        """
        self.m_rh.set_id_file(id, file)

    def set_report_severity_file(self, severity, file):
        """
        Function: set_report_severity_file

        Args:
            severity:
            file:
        """
        self.m_rh.set_severity_file(severity, file)

    # Function: set_report_severity_id_file
    #
    # These methods configure the report handler to direct some or all of its
    # output to the given file descriptor. The ~file~ argument must be a
    # multi-channel descriptor (mcd) or file id compatible with $fdisplay.
    #
    # A FILE descriptor can be associated with reports of
    # the given ~severity~, ~id~, or ~severity-id~ pair.    A FILE associated with
    # a particular ~severity-id~ pair takes precedence over a FILE associated
    # with ~id~, which take precedence over an a FILE associated with a
    # ~severity~, which takes precedence over the default FILE descriptor.
    #
    # When a report is issued and its associated action has the UVM_LOG bit
    # set, the report will be sent to its associated FILE descriptor.
    # The user is responsible for opening and closing these files.

    def set_report_severity_id_file (self, severity, id, file):
        self.m_rh.set_severity_id_file(severity, id, file)

    #----------------------------------------------------------------------------
    # Group: Override Configuration
    #----------------------------------------------------------------------------

    def set_report_severity_override(self, cur_severity, new_severity):
        """
        Args:
            cur_severity:
            new_severity:
        """
        self.m_rh.set_severity_override(cur_severity, new_severity)

    def set_report_severity_id_override(self, cur_severity, id, new_severity):
        """
        These methods provide the ability to upgrade or downgrade a message in
        terms of severity given `severity` and `id`.    An upgrade or downgrade for
        a specific `id` takes precedence over an upgrade or downgrade associated
        with a `severity`.
        Args:
            cur_severity:
            id:
            new_severity:
        """
        self. m_rh.set_severity_id_override(cur_severity, id, new_severity)

    #----------------------------------------------------------------------------
    # Group: Report Handler Configuration
    #----------------------------------------------------------------------------

    def set_report_handler(self, handler):
        """
        Sets the report handler, overwriting the default instance. This allows
        more than one component to share the same report handler.
        Args:
            handler (UVMReportHandler): New report handler for this object.
        """
        self.m_rh = handler

    def get_report_handler(self):
        """
        Returns the underlying report handler to which most reporting tasks
        are delegated.
        Returns:
            UVMReportHandler: Report handler of this object.
        """
        return self.m_rh

    def reset_report_handler(self):
        """
        Resets the underlying report handler to its default settings. This clears
        any settings made with the ~set_report_*~ methods (see below).
        """
        self.m_rh.initialize()





#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.comps import UVMEnv
from uvm.macros import *
from uvm.base import *
from ubus_bus_monitor import ubus_bus_monitor
from ubus_slave_agent import ubus_slave_agent
from ubus_master_agent import ubus_master_agent

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_env
#//
#//------------------------------------------------------------------------------
#class ubus_env extends uvm_env
class ubus_env(UVMEnv):
    #
    #  // Virtual Interface variable
    #  protected virtual interface ubus_if vif
    #
    #
    #  // The following two bits are used to control whether checks and coverage are
    #  // done both in the bus monitor class and the interface.
    #  bit intf_checks_enable = 1; 
    #  bit intf_coverage_enable = 1
    #
    #  // Components of the environment
    #  ubus_bus_monitor bus_monitor
    #  ubus_master_agent masters[]
    #  ubus_slave_agent slaves[]
    #
    #  // Provide implementations of virtual methods such as get_type_name and create
    #  `uvm_component_utils_begin(ubus_env)
    #    `uvm_field_int(has_bus_monitor, UVM_DEFAULT)
    #    `uvm_field_int(num_masters, UVM_DEFAULT)
    #    `uvm_field_int(num_slaves, UVM_DEFAULT)
    #    `uvm_field_int(intf_checks_enable, UVM_DEFAULT)
    #    `uvm_field_int(intf_coverage_enable, UVM_DEFAULT)
    #  `uvm_component_utils_end
    #

    #  // new - constructor
    def __init__(self, name, parent):
        UVMEnv.__init__(self, name, parent)
        self.monitor = None
        self.masters = []
        self.slaves = []
        self.vif = None
        # Control properties
        self.has_bus_monitor = True
        self.num_masters = 1
        self.num_slaves = 1

    #  // build_phase
    def build_phase(self, phase):
        inst_name = ""
        super().build_phase(phase)

        arr = []
        if UVMConfigDb.get(None, "*", "vif", arr):
            uvm_info("GOT_VIF", "vif was received from configDb", UVM_HIGH)
            self.vif = arr[0]

        if self.vif is None:
            self.uvm_report_fatal("NOVIF","virtual interface must be set for: " +
                    self.get_full_name() + ".vif")

        if self.has_bus_monitor is True:
            self.bus_monitor = ubus_bus_monitor.type_id.create("bus_monitor", self)

        arr = []
        if UVMConfigDb.get(self, "", "num_masters", arr) is True:
            self.num_masters = arr[0]

        for i in range(self.num_masters):
            inst_name = sv.sformatf("masters[%0d]", i)
            master = ubus_master_agent.type_id.create(inst_name, self)
            self.masters.append(master)
            UVMConfigDb.set(self, inst_name + ".monitor", "master_id", i)
            UVMConfigDb.set(self, inst_name + ".driver", "master_id", i)

        arr = []
        if UVMConfigDb.get(self, "", "num_slaves", arr) is True:
            self.num_slaves = arr[0]

        for i in range(self.num_slaves):
            inst_name = sv.sformatf("slaves[%0d]", i)
            self.slaves.append(ubus_slave_agent.type_id.create(inst_name, self))


    #
    #  // set_slave_address_map
    def set_slave_address_map(self, slave_name, min_addr, max_addr):
        tmp_slave_monitor = None
        if self.bus_monitor is not None:
            # Set slave address map for bus monitor
            self.bus_monitor.set_slave_configs(slave_name, min_addr, max_addr)

        # Set slave address map for slave monitor
        mon = self.lookup(slave_name + ".monitor")
        tmp_slave_monitor = mon
        if hasattr(mon, 'set_addr_range'):
            tmp_slave_monitor.set_addr_range(min_addr, max_addr)
        #  endfunction : set_slave_address_map

    #
    #  // update_vif_enables
    #  protected task update_vif_enables()
    #    forever begin
    #      @(intf_checks_enable || intf_coverage_enable)
    #      vif.has_checks <= intf_checks_enable
    #      vif.has_coverage <= intf_coverage_enable
    #    end
    #  endtask : update_vif_enables
    #
    #  // implement run task
    #  task run_phase(uvm_phase phase)
    #    fork
    #      update_vif_enables()
    #    join
    #  endtask : run_phase
    #
    #endclass : ubus_env
uvm_component_utils(ubus_env)



#// 
#// -------------------------------------------------------------
#//    Copyright 2010 Cadence.
#//    Copyright 2011 Mentor Graphics Corporation
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//
#
#//
#// TITLE: HDL Paths Checking Test Sequence
#//
#
#//
#// class: uvm_reg_mem_hdl_paths_seq
#//
#// Verify the correctness of HDL paths specified for registers and memories.
#//
#// This sequence is be used to check that the specified backdoor paths
#// are indeed accessible by the simulator.
#// By default, the check is performed for the default design abstraction.
#// If the simulation contains multiple models of the DUT,
#// HDL paths for multiple design abstractions can be checked.
#// 
#// If a path is not accessible by the simulator, it cannot be used for 
#// read/write backdoor accesses. In that case a warning is produced. 
#// A simulator may have finer-grained access permissions such as separate 
#// read or write permissions.
#// These extra access permissions are NOT checked.
#//
#// The test is performed in zero time and
#// does not require any reads/writes to/from the DUT.
#//
#
#class uvm_reg_mem_hdl_paths_seq extends uvm_reg_sequence #(uvm_sequence #(uvm_reg_item));
#    // Variable: abstractions
#    // If set, check the HDL paths for the specified design abstractions.
#    // If empty, check the HDL path for the default design abstraction,
#    // as specified with <uvm_reg_block::set_default_hdl_path()>
#    string abstractions[$];
#    
#    `uvm_object_utils_begin(uvm_reg_mem_hdl_paths_seq)
#        `uvm_field_queue_string(abstractions, UVM_DEFAULT)
#    `uvm_object_utils_end
#    
#    function new(string name="uvm_reg_mem_hdl_paths_seq");
#        super.new(name);
#    endfunction
#
#    virtual task body();
#
#        if (model == null) begin
#            uvm_report_error("uvm_reg_mem_hdl_paths_seq", "Register model handle is null");
#            return;
#        end
#
#       `uvm_info("uvm_reg_mem_hdl_paths_seq",
#                 {"checking HDL paths for all registers/memories in ",
#                  model.get_full_name()}, UVM_LOW);
#
#       if (abstractions.size() == 0)
#          do_block(model, "");
#       else begin
#          foreach (abstractions[i])
#             do_block(model, abstractions[i]);
#       end
#
#        `uvm_info("uvm_reg_mem_hdl_paths_seq", "HDL path validation completed ",UVM_LOW);
#        
#    endtask: body
#
#
#    // Any additional steps required to reset the block
#    // and make it accessible
#    virtual task reset_blk(uvm_reg_block blk);
#    endtask
#
#
#    protected virtual function void do_block(uvm_reg_block blk,
#                                             string        kind);
#        uvm_reg       regs[$];
#        uvm_mem       mems[$];
#
#       `uvm_info("uvm_reg_mem_hdl_paths_seq",
#                 {"Validating HDL paths in ", blk.get_full_name(),
#                  " for ", (kind == "") ? "default" : kind,
#                  " design abstraction"}, UVM_MEDIUM) 
#
#       // Iterate over all registers, checking accesses
#       blk.get_registers(regs, UVM_NO_HIER);
#       foreach (regs[i]) 
#          check_reg(regs[i], kind);
#       
#       blk.get_memories(mems, UVM_NO_HIER);
#       foreach (mems[i]) 
#          check_mem(mems[i], kind);
#    
#       begin
#          uvm_reg_block blks[$];
#          
#          blk.get_blocks(blks);
#          foreach (blks[i]) begin
#             do_block(blks[i], kind);
#          end
#       end
#    endfunction: do_block
#    
#
#    protected virtual function void check_reg(uvm_reg r,
#                                              string kind);
#        uvm_hdl_path_concat paths[$];
#
#	// avoid calling get_full_hdl_path when the register has not path for this abstraction kind
#	if(!r.has_hdl_path(kind))
#		return;
#
#        r.get_full_hdl_path(paths, kind);
#        if (paths.size() == 0) return;
#
#        foreach(paths[p]) begin
#            uvm_hdl_path_concat path=paths[p];
#            foreach (path.slices[j]) begin
#                string p_ = path.slices[j].path;
#                uvm_reg_data_t d;
#                if (!uvm_hdl_read(p_,d))
#                    `uvm_error("uvm_reg_mem_hdl_paths_seq",
#                               $sformatf("HDL path \"%s\" for register \"%s\" is not readable",
#                                         p_, r.get_full_name()));
#                if (!uvm_hdl_check_path(p_))
#                    `uvm_error("uvm_reg_mem_hdl_paths_seq",
#                               $sformatf("HDL path \"%s\" for register \"%s\" is not accessible",
#                                         p_, r.get_full_name()));
#            end
#        end
#    endfunction
# 
#
#    protected virtual function void check_mem(uvm_mem m,
#                                              string kind);
#        uvm_hdl_path_concat paths[$];
#
#	// avoid calling get_full_hdl_path when the register has not path for this abstraction kind
#	if(!m.has_hdl_path(kind))
#		return;
#
#        m.get_full_hdl_path(paths, kind);
#        if (paths.size() == 0) return;
#
#        foreach(paths[p]) begin
#            uvm_hdl_path_concat path=paths[p];
#            foreach (path.slices[j]) 
#            begin
#                string p_ = path.slices[j].path;
#                if(!uvm_hdl_check_path(p_))
#                    `uvm_error("uvm_reg_mem_hdl_paths_seq",
#                               $sformatf("HDL path \"%s\" for memory \"%s\" is not accessible",
#                                         p_, m.get_full_name()));
#            end
#        end
#    endfunction 
#endclass: uvm_reg_mem_hdl_paths_seq



#// 
#// -------------------------------------------------------------
#//    Copyright 2010-2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm.seq.uvm_sequence_item import *
from uvm.macros import *
from uvm.base import sv
from enum import Enum, auto

#typedef enum {USB_TLM_TOKEN, USB_TLM_DATA, USB_TLM_HANDSHAKE} usb_tlm_phase
USB_TLM_TOKEN = 0
USB_TLM_DATA = 1
USB_TLM_HANDSHAKE = 2


class status_e(Enum):
    ACK = auto()
    NAK = auto()
    STALL = auto()
    NYET = auto()


class usb_xfer(UVMSequenceItem):

    IN = 10
    OUT = 20

    ACK = status_e.ACK
    NAK = status_e.NAK
    STALL = status_e.STALL
    NYET = status_e.NYET

    #   rand kind_e   kind;  
    #   rand byte      data[]
    #   rand status_e  status


    def __init__(self, name="usb_xfer"):
        super().__init__(name)
        self.addr = 0
        self.rand('addr', range(0, (1 << 7) - 1))
        self.endp = 0x0
        self.rand('endp', range(0, (1 << 4) - 1))
        self.kind = usb_xfer.IN
        self.rand('kind', [usb_xfer.IN, usb_xfer.OUT])
        self.data = []
        self.status = usb_xfer.ACK

    def convert2string(self):
        convert2string = sv.sformatf("dev=%h.%h kind=%s hndsk=%s data=h",
            self.addr, self.endp, int(self.kind), self.status.name)
        ndata = len(self.data)

        if ndata == 0:
            return convert2string + "[]"
        if ndata == 1:
            return sv.sformatf("%s0x%h", convert2string, self.data[0])
        if ndata == 2:
            return sv.sformatf("%s0x%h 0x%h", convert2string, self.data[0],
                    self.data[1])
        if ndata == 3:
            return sv.sformatf("%s0x%h 0x%h 0x%h", convert2string,
                self.data[0], self.data[1], self.data[2])

        return sv.sformatf("%s0x%h 0x%h .. 0x%h (%0d bytes)", convert2string,
            self.data[0], self.data[1], self.data[-1],
            len(self.data))

uvm_object_utils_begin(usb_xfer)
#     `uvm_field_enum(kind_e, kind, UVM_ALL_ON | UVM_NOPACK)
#     `uvm_field_array_int(data, UVM_ALL_ON | UVM_NOPACK)
#     `uvm_field_enum(status_e, status, UVM_ALL_ON | UVM_NOPACK)
uvm_object_utils_end(usb_xfer)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer, Combine
from cocotb.utils import get_sim_time

from uvm.base.uvm_debug import UVMDebug
from uvm.base.sv import sv
from uvm.base.uvm_globals import run_test, uvm_sim_time
from uvm.base.uvm_object_globals import UVM_MEDIUM, UVM_HIGH
from uvm.base.uvm_phase import UVMPhase
from uvm.seq import UVMSequence, UVMSequenceItem, UVMSequencer
from uvm.macros import *
from uvm.comps import UVMDriver, UVMEnv

NUM_SEQS = 16
NUM_LOOPS = 1

start_times = {}
finish_times = {}


class MyReq(UVMSequenceItem):

    def __init__(self, name="my_seq_item"):
        super().__init__(name)


uvm_object_utils(MyReq)


class my_driver(UVMDriver):

    def __init__(self, name, parent):
        UVMDriver.__init__(self, name, parent)
        self.data_array = [0] * 512  # int data_array[511:0]
        self.nitems = 0
        self.recording_detail = UVM_HIGH

    
    async def main_phase(self, phase):
        # req = None
        # rsp = None

        while True:
            uvm_info("DRIVER", "my_driver getting the req now..",
                UVM_MEDIUM)
            qreq = []
            await self.seq_item_port.get_next_item(qreq)
            async def fork_task():
                await Timer(100, "NS")
            cocotb.fork(fork_task())
            self.nitems += 1
            self.seq_item_port.item_done()
            #self.end_tr(req)

uvm_component_utils(my_driver)


class sequenceA(UVMSequence):

    g_my_id = 1

    def __init__(self, name=""):
        UVMSequence.__init__(self, name)
        self.my_id = sequenceA.g_my_id
        sequenceA.g_my_id += 1
        self.id = -1

    async def body(self):
        uvm_info("sequenceA", "Starting sequence with ID {}".format(self.id), UVM_MEDIUM)
        req = MyReq()
        start_times[self.id] = uvm_sim_time()
        await self.start_item(req)
        await self.finish_item(req)
        finish_times[self.id] = uvm_sim_time()

uvm_object_utils(sequenceA)


class env(UVMEnv):


    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.drv = None
        self.sqr = None
        self.error = False

    def build_phase(self, phase):
        super().build_phase(phase)
        self.sqr = UVMSequencer("sequencer", self)

        # create and connect driver
        self.drv = my_driver("my_driver", self)
        uvm_info("ENV", "Connecting ports now here", UVM_MEDIUM)
        self.drv.seq_item_port.connect(self.sqr.seq_item_export)

    
    async def main_phase(self, phase):
        phase.raise_objection(self)
        uvm_info("ENV", 'Before 1000 x await Timer(0)', UVM_LOW)
        for _ in range(1000):
            await Timer(0)
        uvm_info("ENV", 'main_phase logic starting', UVM_LOW)
        fork_procs = []
        for i in range(NUM_SEQS):
            the_sequence = sequenceA("sequence_" + str(i))
            the_sequence.id = i
            fork_procs.append(cocotb.fork(the_sequence.start(self.sqr, None)))
        sv.fork_join(fork_procs)
        # join_list = list(map(lambda t: t.join(), fork_procs))
        await Timer(100, "NS")
        phase.drop_objection(self)

    def check_phase(self, phase):
        if self.drv.nitems != (NUM_SEQS * NUM_LOOPS):
            self.error = True
            self.uvm_report_error('env', sv.sformatf("nitems: Exp %0d, Got %0d",
                NUM_SEQS * NUM_LOOPS, self.drv.nitems))

    def report_phase(self, phase):
        rpt = "Start times:\n" + str(start_times)
        rpt += "\nFinish times:\n" + str(finish_times)
        uvm_info("ENV", rpt, UVM_LOW)


uvm_component_utils(env)


@cocotb.test()
async def seq_fork_module_top(dut):
    print("Using simulator " + cocotb.SIM_NAME)
    UVMPhase.m_phase_trace = True
    e = env("env_name", parent=None)
    uvm_info("top","In top initial block", UVM_MEDIUM)
    await run_test()
    sim_time = sim_time = get_sim_time('ns')

    if sim_time == 0:
        raise Exception('Sim time has not progressed')
    else:
        if sim_time < 50.0:
            msg = "Exp: > 50, got: " + str(sim_time)
            raise Exception('Wrong sim time at the end: ' + msg)
    if e.error is True:
        raise Exception('Env had errors')



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from ubus_master_seq_lib import ubus_base_sequence, read_byte_seq, write_byte_seq
from uvm.macros import uvm_do_with, uvm_error, uvm_info, uvm_object_utils
from uvm.base import sv, UVM_MEDIUM

#//------------------------------------------------------------------------------
#//
#// SEQUENCE: incr_read_byte_seq
#//
#//------------------------------------------------------------------------------

#class incr_read_byte_seq extends ubus_base_sequence;
#
#  function new(string name="incr_read_byte_seq");
#    super.new(name);
#  endfunction : new
#
#  `uvm_object_utils(incr_read_byte_seq)
#
#  read_byte_seq read_byte_seq0;
#
#  rand int unsigned count;
#    constraint count_ct { (count < 20); }
#  rand bit [15:0] start_address;
#  rand int unsigned incr_transmit_del = 0;
#    constraint transmit_del_ct { (incr_transmit_del <= 10); }
# 
#  virtual task body();
#    `uvm_info(get_type_name(),
#      $sformatf("%s starting with count = %0d", 
#      get_sequence_path(), count), UVM_MEDIUM);
#    repeat(count) begin : repeat_block
#      `uvm_do_with(read_byte_seq0,
#        { read_byte_seq0.start_addr == start_address;
#          read_byte_seq0.transmit_del == incr_transmit_del; } )
#      start_address++;
#    end : repeat_block
#  endtask : body
# 
#endclass : incr_read_byte_seq


#//------------------------------------------------------------------------------
#//
#// SEQUENCE: incr_write_byte_seq
#//
#//------------------------------------------------------------------------------
#
#class incr_write_byte_seq extends ubus_base_sequence;
#
#  function new(string name="incr_write_byte_seq");
#    super.new(name);
#  endfunction : new
#
#  `uvm_object_utils(incr_write_byte_seq)
#    
#  write_byte_seq write_byte_seq0;
#
#  rand int unsigned count;
#    constraint count_ct { (count < 20); }
#  rand bit [15:0] start_address;
#  rand int unsigned incr_transmit_del = 0;
#    constraint transmit_del_ct { (incr_transmit_del <= 10); }
#
#  virtual task body();
#    `uvm_info(get_type_name(),
#      $sformatf("%s starting with count = %0d",
#      get_sequence_path(), count), UVM_MEDIUM);
#    repeat(count) begin : repeat_block
#      `uvm_do_with(write_byte_seq0,
#        { write_byte_seq0.start_addr == start_address;
#          write_byte_seq0.transmit_del == incr_transmit_del; } )
#      start_address++;
#    end : repeat_block
#  endtask : body
#
#endclass : incr_write_byte_seq

#//------------------------------------------------------------------------------
#//
#// SEQUENCE: incr_read_write_read_seq
#//
#//------------------------------------------------------------------------------

#class incr_read_write_read_seq extends ubus_base_sequence;
#
#  function new(string name="incr_read_write_read_seq");
#    super.new(name);
#  endfunction : new
#
#  `uvm_object_utils(incr_read_write_read_seq)
#
#  incr_read_byte_seq  read0;
#  incr_write_byte_seq write0;
#
#  virtual task body();
#    `uvm_info(get_type_name(),
#      $sformatf("%s starting sequence",
#      get_sequence_path()), UVM_MEDIUM);
#    `uvm_do(read0)
#    `uvm_do(write0)
#    `uvm_do(read0)
#  endtask : body
#
#endclass : incr_read_write_read_seq


#//------------------------------------------------------------------------------
#//
#// SEQUENCE: r8_w8_r4_w4_seq
#//
#//------------------------------------------------------------------------------
#class r8_w8_r4_w4_seq extends ubus_base_sequence;
#
#  function new(string name="r8_w8_r4_w4_seq");
#    super.new(name);
#  endfunction : new
#
#  `uvm_object_utils(r8_w8_r4_w4_seq)
#
#  read_word_seq read_word_seq0;
#  read_double_word_seq read_double_word_seq0;
#  write_word_seq write_word_seq0;
#  write_double_word_seq write_double_word_seq0;
#
#  rand bit [15:0] start_address;
#
#  constraint start_address_ct { (start_address == 16'h4000); }
#
#  virtual task body();
#    `uvm_info(get_type_name(),
#      $sformatf("%s starting...",
#      get_sequence_path()), UVM_MEDIUM);
#    `uvm_do_with(read_double_word_seq0, 
#      { read_double_word_seq0.start_addr == start_address;
#        read_double_word_seq0.transmit_del == 2; } )
#    `uvm_do_with(write_double_word_seq0,
#      { write_double_word_seq0.start_addr == start_address;
#        write_double_word_seq0.transmit_del == 4; } )
#    start_address = start_address + 8;
#    `uvm_do_with(read_word_seq0,
#      { read_word_seq0.start_addr == start_address;
#        read_word_seq0.transmit_del == 6; } )
#    `uvm_do_with(write_word_seq0,
#      { write_word_seq0.start_addr == start_address;
#        write_word_seq0.transmit_del == 8; } )
#  endtask : body
#
#endclass : r8_w8_r4_w4_seq


#//------------------------------------------------------------------------------
#//
#// SEQUENCE: read_modify_write_seq
#//
#//------------------------------------------------------------------------------
#class read_modify_write_seq extends ubus_base_sequence;

class read_modify_write_seq(ubus_base_sequence):

    def __init__(self, name="read_modify_write_seq"):
        ubus_base_sequence.__init__(self, name)
        self.read_byte_seq0 = None  # read_byte_seq
        self.write_byte_seq0 = None  # write_byte_seq
        self.addr_check = 0
        self.m_data0_check = 0
        self.test_pass = False

    
    async def body(self):
        uvm_info(self.get_type_name(), sv.sformatf("%s starting...",
            self.get_sequence_path()), UVM_MEDIUM)
        # READ A RANDOM LOCATION
        self.read_byte_seq0 = read_byte_seq("read_byte_seq")
        await uvm_do_with(self, self.read_byte_seq0, {})
        self.addr_check = self.read_byte_seq0.rsp.addr
        self.m_data0_check = self.read_byte_seq0.rsp.data[0] + 1
        
        # WRITE MODIFIED READ DATA
        self.write_byte_seq0 = write_byte_seq("write_byte_seq")
        self.write_byte_seq0.start_addr = self.addr_check
        self.write_byte_seq0.data0 = self.m_data0_check
        await uvm_do_with(self, self.write_byte_seq0, {})
        #      { write_byte_seq0.start_addr == addr_check;
        #        write_byte_seq0.data0 == m_data0_check; } )

        self.m_data0_check = write_byte_seq.last_data
    
        #    // READ MODIFIED WRITE DATA
        self.read_byte_seq0.start_addr = self.addr_check
        await uvm_do_with(self, self.read_byte_seq0,
            lambda start_addr: start_addr == self.addr_check)

        data0_got = int(self.read_byte_seq0.rsp.data[0])
        if self.m_data0_check != data0_got:
            uvm_error(self.get_type_name(),
                sv.sformatf("%s Read Modify Write Read error!\n\tADDR: %h, EXP: %h, ACT: %h", 
                    self.get_sequence_path(),
                    self.addr_check, self.m_data0_check, data0_got))
        else:
            self.test_pass = True


uvm_object_utils(read_modify_write_seq)

#//------------------------------------------------------------------------------
#//
#// SEQUENCE: loop_read_modify_write_seq
#//
#//------------------------------------------------------------------------------
#
#class loop_read_modify_write_seq extends ubus_base_sequence;
#
#  int itr;
#
#  function new(string name="loop_read_modify_write_seq");
#    super.new(name);
#  endfunction : new
#
#  `uvm_object_utils(loop_read_modify_write_seq)
#
#  read_modify_write_seq rmw_seq;
#
#  virtual task body();
#    void'(uvm_config_db#(int)::get(null,get_full_name(),"itr", itr));
#    `uvm_info(get_type_name(),
#      $sformatf("%s starting...itr = %0d",
#      get_sequence_path(),itr), UVM_NONE);
#    for(int i = 0; i < itr; i++) begin
#      `uvm_do(rmw_seq)
#    end
#  endtask : body
#
#endclass : loop_read_modify_write_seq
#



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer

from ..base.uvm_callback import UVMCallback, UVMCallbackIter, UVMCallbacks
from ..base.uvm_globals import uvm_zero_delay
from ..macros import uvm_object_utils

#//------------------------------------------------------------------------------
#// Title: Register Callbacks
#//
#// This section defines the base class used for all register callback
#// extensions. It also includes pre-defined callback extensions for use on
#// read-only and write-only registers.
#//------------------------------------------------------------------------------

#//------------------------------------------------------------------------------
#// Class: uvm_reg_cbs
#//
#// Facade class for field, register, memory and backdoor
#// access callback methods.
#//------------------------------------------------------------------------------


class UVMRegCbs(UVMCallback):

    def __init__(self, name="uvm_reg_cbs"):
        UVMCallback.__init__(self, name)

    async def pre_write(self, rw):
        """
           Task: pre_write

           Called before a write operation.

           All registered `pre_write` callback methods are invoked after the
           invocation of the `pre_write` method of associated object (`uvm_reg`,
           `uvm_reg_field`, `uvm_mem`, or `uvm_reg_backdoor`). If the element being
           written is a `uvm_reg`, all `pre_write` callback methods are invoked
           before the contained `uvm_reg_fields`.

           Backdoor - <uvm_reg_backdoor::pre_write>,
                      <uvm_reg_cbs::pre_write> cbs for backdoor.

           Register - <uvm_reg::pre_write>,
                      <uvm_reg_cbs::pre_write> cbs for reg,
                      then foreach field:
                        <uvm_reg_field::pre_write>,
                        <uvm_reg_cbs::pre_write> cbs for field

           RegField - <uvm_reg_field::pre_write>,
                      <uvm_reg_cbs::pre_write> cbs for field

           Memory   - <uvm_mem::pre_write>,
                      <uvm_reg_cbs::pre_write> cbs for mem

           The `rw` argument holds information about the operation.

           - Modifying the `value` modifies the actual value written.

           - For memories, modifying the `offset` modifies the offset
             used in the operation.

           - For non-backdoor operations, modifying the access `path` or
             address `map` modifies the actual path or map used in the
             operation.

           If the ~rw.status~ is modified to anything other than `UVM_IS_OK`,
           the operation is aborted.

           See `uvm_reg_item` for details on `rw` information.

          virtual task pre_write(uvm_reg_item rw); endtask
        Args:
            rw:
        """
        await uvm_zero_delay()

    async def post_write(self, rw):
        """
           Task: post_write

           Called after a write operation.

           All registered `post_write` callback methods are invoked before the
           invocation of the `post_write` method of the associated object (`uvm_reg`,
           `uvm_reg_field`, `uvm_mem`, or `uvm_reg_backdoor`). If the element being
           written is a `uvm_reg`, all `post_write` callback methods are invoked
           before the contained `uvm_reg_fields`.

           Summary of callback order:

           Backdoor - <uvm_reg_cbs::post_write> cbs for backdoor,
                      <uvm_reg_backdoor::post_write>

           Register - <uvm_reg_cbs::post_write> cbs for reg,
                      <uvm_reg::post_write>,
                      then foreach field:
                        <uvm_reg_cbs::post_write> cbs for field,
                        <uvm_reg_field::post_read>

           RegField - <uvm_reg_cbs::post_write> cbs for field,
                      <uvm_reg_field::post_write>

           Memory   - <uvm_reg_cbs::post_write> cbs for mem,
                      <uvm_mem::post_write>

           The `rw` argument holds information about the operation.

           - Modifying the `status` member modifies the returned status.

           - Modifying the `value` or `offset` members has no effect, as
             the operation has already completed.

           See `uvm_reg_item` for details on `rw` information.

          virtual task post_write(uvm_reg_item rw); endtask
        Args:
            rw:
        """
        await uvm_zero_delay()

    async def pre_read(self, rw):
        """
           Task: pre_read

           Callback called before a read operation.

           All registered `pre_read` callback methods are invoked after the
           invocation of the `pre_read` method of associated object (`uvm_reg`,
           `uvm_reg_field`, `uvm_mem`, or `uvm_reg_backdoor`). If the element being
           read is a `uvm_reg`, all `pre_read` callback methods are invoked before
           the contained `uvm_reg_fields`.

           Backdoor - <uvm_reg_backdoor::pre_read>,
                      <uvm_reg_cbs::pre_read> cbs for backdoor

           Register - <uvm_reg::pre_read>,
                      <uvm_reg_cbs::pre_read> cbs for reg,
                      then foreach field:
                        <uvm_reg_field::pre_read>,
                        <uvm_reg_cbs::pre_read> cbs for field

           RegField - <uvm_reg_field::pre_read>,
                      <uvm_reg_cbs::pre_read> cbs for field

           Memory   - <uvm_mem::pre_read>,
                      <uvm_reg_cbs::pre_read> cbs for mem

           The `rw` argument holds information about the operation.

           - The `value` member of `rw` is not used has no effect if modified.

           - For memories, modifying the `offset` modifies the offset
             used in the operation.

           - For non-backdoor operations, modifying the access `path` or
             address `map` modifies the actual path or map used in the
             operation.

           If the ~rw.status~ is modified to anything other than `UVM_IS_OK`,
           the operation is aborted.

           See `uvm_reg_item` for details on `rw` information.

          virtual task pre_read(uvm_reg_item rw); endtask
        Args:
            rw:
        """
        await uvm_zero_delay()


    def post_read(self, rw):
        """
           Task: post_read

           Callback called after a read operation.

           All registered `post_read` callback methods are invoked before the
           invocation of the `post_read` method of the associated object (`uvm_reg`,
           `uvm_reg_field`, `uvm_mem`, or `uvm_reg_backdoor`). If the element being read
           is a `uvm_reg`, all `post_read` callback methods are invoked before the
           contained `uvm_reg_fields`.

           Backdoor - <uvm_reg_cbs::post_read> cbs for backdoor,
                      <uvm_reg_backdoor::post_read>

           Register - <uvm_reg_cbs::post_read> cbs for reg,
                      <uvm_reg::post_read>,
                      then foreach field:
                        <uvm_reg_cbs::post_read> cbs for field,
                        <uvm_reg_field::post_read>

           RegField - <uvm_reg_cbs::post_read> cbs for field,
                      <uvm_reg_field::post_read>

           Memory   - <uvm_reg_cbs::post_read> cbs for mem,
                      <uvm_mem::post_read>

           The `rw` argument holds information about the operation.

           - Modifying the readback `value` or `status` modifies the actual
             returned value and status.

           - Modifying the `value` or `offset` members has no effect, as
             the operation has already completed.

           See `uvm_reg_item` for details on `rw` information.

          virtual task post_read(uvm_reg_item rw); endtask
        Args:
            rw:
        """
        pass


    #   // Task: post_predict
    #   //
    #   // Called by the <uvm_reg_field::predict()> method
    #   // after a successful UVM_PREDICT_READ or UVM_PREDICT_WRITE prediction.
    #   //
    #   // ~previous~ is the previous value in the mirror and
    #   // ~value~ is the latest predicted value. Any change to ~value~ will
    #   // modify the predicted mirror value.
    #   //
    #   virtual function void post_predict(input uvm_reg_field  fld,
    #                                      input uvm_reg_data_t previous,
    #                                      inout uvm_reg_data_t value,
    #                                      input uvm_predict_e  kind,
    #                                      input uvm_path_e     path,
    #                                      input uvm_reg_map    map)
    #   endfunction


    def encode(self, data):
        """
           Function: encode

           Data encoder

           The registered callback methods are invoked in order of registration
           after all the `pre_write` methods have been called.
           The encoded data is passed through each invocation in sequence.
           This allows the `pre_write` methods to deal with clear-text data.

           By default, the data is not modified.

        Args:
            data:
        """
        pass


    def decode(self, data):
        """
           Function: decode

           Data decode

           The registered callback methods are invoked in ~reverse order~
           of registration before all the `post_read` methods are called.
           The decoded data is passed through each invocation in sequence.
           This allows the `post_read` methods to deal with clear-text data.

           The reversal of the invocation order is to allow the decoding
           of the data to be performed in the opposite order of the encoding
           with both operations specified in the same callback extension.

           By default, the data is not modified.

        Args:
            data:
        """
        pass


#//------------------
#// Section: Typedefs
#//------------------


#// Type: uvm_reg_cb
#//
#// Convenience callback type declaration for registers
#//
#// Use this declaration to register the register callbacks rather than
#// the more verbose parameterized class
#//
#typedef uvm_callbacks#(uvm_reg, uvm_reg_cbs) uvm_reg_cb
UVMRegCb = UVMCallbacks

#// Type: uvm_reg_cb_iter
#//
#// Convenience callback iterator type declaration for registers
#//
#// Use this declaration to iterate over registered register callbacks
#// rather than the more verbose parameterized class
#//
#typedef uvm_callback_iter#(uvm_reg, uvm_reg_cbs) uvm_reg_cb_iter
UVMRegCbIter = UVMCallbackIter



#// Type: uvm_reg_bd_cb
#//
#// Convenience callback type declaration for backdoor
#//
#// Use this declaration to register register backdoor callbacks rather than
#// the more verbose parameterized class
#//
#typedef uvm_callbacks#(uvm_reg_backdoor, uvm_reg_cbs) uvm_reg_bd_cb


#// Type: uvm_reg_bd_cb_iter
#// Convenience callback iterator type declaration for backdoor
#//
#// Use this declaration to iterate over registered register backdoor callbacks
#// rather than the more verbose parameterized class
#//
#typedef uvm_callback_iter#(uvm_reg_backdoor, uvm_reg_cbs) uvm_reg_bd_cb_iter


#// Type: uvm_mem_cb
#//
#// Convenience callback type declaration for memories
#//
#// Use this declaration to register memory callbacks rather than
#// the more verbose parameterized class
#//
#typedef uvm_callbacks#(uvm_mem, uvm_reg_cbs) uvm_mem_cb

#
#// Type: uvm_mem_cb_iter
#//
#// Convenience callback iterator type declaration for memories
#//
#// Use this declaration to iterate over registered memory callbacks
#// rather than the more verbose parameterized class
#//
#typedef uvm_callback_iter#(uvm_mem, uvm_reg_cbs) uvm_mem_cb_iter
UVMMemCbIter = UVMCallbackIter

#
#// Type: uvm_reg_field_cb
#//
#// Convenience callback type declaration for fields
#//
#// Use this declaration to register field callbacks rather than
#// the more verbose parameterized class
#//
#typedef uvm_callbacks#(uvm_reg_field, uvm_reg_cbs) uvm_reg_field_cb
UVMRegFieldCb = UVMCallbacks

#
#// Type: uvm_reg_field_cb_iter
#//
#// Convenience callback iterator type declaration for fields
#//
#// Use this declaration to iterate over registered field callbacks
#// rather than the more verbose parameterized class
#//
#typedef uvm_callback_iter#(uvm_reg_field, uvm_reg_cbs) uvm_reg_field_cb_iter
UVMRegFieldCbIter = UVMCallbackIter


#//-----------------------------
#// Group: Predefined Extensions
#//-----------------------------
#
#//------------------------------------------------------------------------------
#// Class: uvm_reg_read_only_cbs
#//
#// Pre-defined register callback method for read-only registers
#// that will issue an error if a write() operation is attempted.
#//
#//------------------------------------------------------------------------------
#

#class uvm_reg_read_only_cbs extends uvm_reg_cbs
class UVMRegReadOnlyCbs(UVMRegCbs):

    def __init__(self, name="uvm_reg_read_only_cbs"):
        UVMRegCbs.__init__(self, name)

    #
    #
    #   // Function: pre_write
    #   //
    #   // Produces an error message and sets status to <UVM_NOT_OK>.
    #   //
    #   virtual task pre_write(uvm_reg_item rw)
    #      string name = rw.element.get_full_name()
    #
    #      if (rw.status != UVM_IS_OK)
    #         return
    #
    #      if (rw.element_kind == UVM_FIELD) begin
    #         uvm_reg_field fld
    #         uvm_reg rg
    #         $cast(fld, rw.element)
    #         rg = fld.get_parent()
    #         name = rg.get_full_name()
    #      end
    #
    #      `uvm_error("UVM/REG/READONLY",
    #                 {name, " is read-only. Cannot call write() method."})
    #
    #      rw.status = UVM_NOT_OK
    #   endtask
    #

    m_me = None
    @classmethod
    def get(cls):
        if cls.m_me is None:
            cls.m_me = UVMRegReadOnlyCbs()
        return cls.m_me


    #   // Function: add
    #   //
    #   // Add this callback to the specified register and its contained fields.
    #   //
    @classmethod
    def add(cls, rg):
        flds = []  # uvm_reg_field [$]
        UVMRegCb.add(rg, cls.get())
        rg.get_fields(flds)
        for i in range(len(flds)):
            UVMRegFieldCb.add(flds[i], cls.get())


    #   // Function: remove
    #   //
    #   // Remove this callback from the specified register and its contained fields.
    #   //
    #   static function void remove(uvm_reg rg)
    #      uvm_reg_cb_iter cbs = new(rg)
    #      uvm_reg_field flds[$]
    #
    #      void'(cbs.first())
    #      while (cbs.get_cb() != get()) begin
    #         if (cbs.get_cb() == null)
    #            return
    #         void'(cbs.next())
    #      end
    #      uvm_reg_cb::delete(rg, get())
    #      rg.get_fields(flds)
    #      foreach (flds[i]) begin
    #         uvm_reg_field_cb::delete(flds[i], get())
    #      end
    #   endfunction
    #endclass
uvm_object_utils(UVMRegReadOnlyCbs)

#//------------------------------------------------------------------------------
#// Class: UVMRegWriteOnlyCbs
#//
#// Pre-defined register callback method for write-only registers
#// that will issue an error if a read() operation is attempted.
#//
#//------------------------------------------------------------------------------


class UVMRegWriteOnlyCbs(UVMRegCbs):

    def __init__(self, name="UVMRegWriteOnlyCbs"):
        super().__init__(name)


    #   // Function: pre_read
    #   //
    #   // Produces an error message and sets status to <UVM_NOT_OK>.
    #   //
    #   virtual task pre_read(uvm_reg_item rw)
    #      string name = rw.element.get_full_name()
    #
    #      if (rw.status != UVM_IS_OK)
    #         return
    #
    #      if (rw.element_kind == UVM_FIELD) begin
    #         uvm_reg_field fld
    #         uvm_reg rg
    #         $cast(fld, rw.element)
    #         rg = fld.get_parent()
    #         name = rg.get_full_name()
    #      end
    #
    #      `uvm_error("UVM/REG/WRTEONLY",
    #                 {name, " is write-only. Cannot call read() method."})
    #
    #      rw.status = UVM_NOT_OK
    #   endtask


    m_me = None

    @classmethod
    def get(cls):
        if cls.m_me is None:
            cls.m_me = UVMRegWriteOnlyCbs()
        return cls.m_me


    #   // Function: add
    #   //
    #   // Add this callback to the specified register and its contained fields.
    #   //
    @classmethod
    def add(cls, rg):
        flds = []  # uvm_reg_field [$]
        UVMRegCb.add(rg, cls.get())
        rg.get_fields(flds)
        for i in range(len(flds)):
            UVMRegFieldCb.add(flds[i], cls.get())


    #   // Function: remove
    #   //
    #   // Remove this callback from the specified register and its contained fields.
    #   //
    #   static function void remove(uvm_reg rg)
    #      uvm_reg_cb_iter cbs = new(rg)
    #      uvm_reg_field flds[$]
    #
    #      void'(cbs.first())
    #      while (cbs.get_cb() != get()) begin
    #         if (cbs.get_cb() == null)
    #            return
    #         void'(cbs.next())
    #      end
    #      uvm_reg_cb::delete(rg, get())
    #      rg.get_fields(flds)
    #      foreach (flds[i]) begin
    #         uvm_reg_field_cb::delete(flds[i], get())
    #      end
    #   endfunction
    #endclass

uvm_object_utils(UVMRegWriteOnlyCbs)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
# `include "comps/uvm_pair.svh"
# `include "comps/uvm_policies.svh"
# `include "comps/uvm_in_order_comparator.svh"
# `include "comps/uvm_algorithmic_comparator.svh"
# `include "comps/uvm_random_stimulus.svh"
# `include "comps/uvm_subscriber.svh"
#
# `include "comps/uvm_monitor.svh"
# `include "comps/uvm_driver.svh"
# `include "comps/uvm_push_driver.svh"
# `include "comps/uvm_scoreboard.svh" 
# `include "comps/uvm_agent.svh"
# `include "comps/uvm_env.svh"
# `include "comps/uvm_test.svh"




import unittest

from uvm.base.uvm_resource_db import (UVMResourceDb, ResourceDbClassFactory)


class TestUVMResourceDb(unittest.TestCase):


    def test_create_class(self):
        DbInt = ResourceDbClassFactory('DbInt', ['_id', 'val2'], int)
        DbStr = ResourceDbClassFactory('DbStr', ['_id', 'val2'], str)
        my_db_int = DbInt(_id=123, val2='name')
        my_db_str = DbStr(_id=123, val2='name')
        self.assertEqual(my_db_int.T, int)
        self.assertEqual(my_db_str.T, str)
        my_db_int.set('*', 'value', 123)
        my_db_str.set('*', 'value', 'str_value')
        rsc_int = my_db_int.get_by_name('*', 'value')
        rsc_str = my_db_str.get_by_name('*', 'value')
        self.assertEqual(rsc_int.read(), 123)
        # TODO self.assertEqual(rsc_str.read(), 'str_value')



#// -------------------------------------------------------------
#//    Copyright 2010 Synopsys, Inc.
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2011 Mentor Graphics Corporation
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from typing import List

from .uvm_reg import UVMReg
from .uvm_reg_model import *
from .uvm_reg_item import UVMRegItem
from .uvm_reg_sequence import UVMRegFrontdoor
from ..base.uvm_resource_db import UVMResourceDb
from ..base.sv import sv
from ..macros import uvm_fatal, uvm_error, uvm_warning
from ..base.uvm_globals import uvm_check_output_args


#//-----------------------------------------------------------------
#// CLASS: uvm_reg_indirect_data
#// Indirect data access abstraction class
#//
#// Models the behavior of a register used to indirectly access
#// a register array, indexed by a second ~address~ register.
#//
#// This class should not be instantiated directly.
#// A type-specific class extension should be used to
#// provide a factory-enabled constructor and specify the
#// ~n_bits~ and coverage models.
#//-----------------------------------------------------------------


class UVMRegIndirectData(UVMReg):


    def __init__(self, name="uvm_reg_indirect", n_bits=0, has_cover=False):
        """         
           Function: new
           Create an instance of this class
          
           Should not be called directly,
           other than via super.new().
           The value of `n_bits` must match the number of bits
           in the indirect register array.
          function new(string name = "uvm_reg_indirect",
                       int unsigned n_bits,
                       int has_cover)
        Args:
            name: 
            n_bits: 
            has_cover: 
        """
        UVMReg.__init__(self, name,n_bits,has_cover)
        self.m_idx = None
        self.m_tbl = []
        #endfunction: new

    def build(self):
        pass


    def configure(self, idx, reg_a, blk_parent, regfile_parent=None):
        """         
           Function: configure
           Configure the indirect data register.
          
           The `idx` register specifies the index,
           in the `reg_a` register array, of the register to access.
           The `idx` must be written to first.
           A read or write operation to this register will subsequently
           read or write the indexed register in the register array.
          
           The number of bits in each register in the register array must be
           equal to `n_bits` of this register.
          
           See <uvm_reg::configure()> for the remaining arguments.
          function void configure (uvm_reg idx,
                                   uvm_reg reg_a[],
                                   uvm_reg_block blk_parent,
                                   uvm_reg_file regfile_parent = None)
        Args:
            idx: 
            reg_a: 
            blk_parent: 
            regfile_parent: 
        """
        super().configure(blk_parent, regfile_parent, "")
        self.m_idx = idx
        self.m_tbl = reg_a

        # Not testable using pre-defined sequences
        UVMResourceDb.set("REG::" + self.get_full_name(), "NO_REG_TESTS", 1)

        # Add a frontdoor to each indirectly-accessed register
        # for every address map this register is in.
        # foreach (m_maps[map]):
        for _map in self.m_maps:
            self.add_frontdoors(_map)


    def add_map(self, _map):
        """         
          /*local*/ virtual function void add_map(uvm_reg_map map)
        Args:
            _map: 
        """
        super().add_map(_map)
        self.add_frontdoors(_map)
    #   endfunction


    def add_frontdoors(self, _map):
        """         
          local function void add_frontdoors(uvm_reg_map map)
        Args:
            _map: 
        """
        for i in range(len(self.m_tbl)):
            fd = None  # uvm_reg_indirect_ftdr_seq
            if self.m_tbl[i] is None:
                uvm_error(self.get_full_name(),
                         sv.sformatf("Indirect register #%0d is None", i))
                continue
            fd = uvm_reg_indirect_ftdr_seq(self.m_idx, i, self)
            if self.m_tbl[i].is_in_map(_map):
                self.m_tbl[i].set_frontdoor(fd, _map)
            else:
                _map.add_reg(self.m_tbl[i], -1, "RW", 1, fd)


    def do_predict(self, rw, kind=UVM_PREDICT_DIRECT, be=-1):
        """         
          virtual function void do_predict (uvm_reg_item      rw,
                                            uvm_predict_e     kind = UVM_PREDICT_DIRECT,
                                            uvm_reg_byte_en_t be = -1)
        Args:
            rw: 
            kind: 
            UVM_PREDICT_DIRECT: 
            be: 
        """
        if self.m_idx.get() >= len(self.m_tbl):
            uvm_error(self.get_full_name(), sv.sformatf(
                "Address register %s has a value (%0d) greater than the maximum indirect register array size (%0d)",
                self.m_idx.get_full_name(), self.m_idx.get(), len(self.m_tbl)))
            rw.status = UVM_NOT_OK
            return

        # NOTE limit to 2**32 registers
        idx = self.m_idx.get()
        self.m_tbl[idx].do_predict(rw, kind, be)

    def get_local_map(self, _map, caller=""):
        """         

          virtual function uvm_reg_map get_local_map(uvm_reg_map map, string caller="")
        Args:
            _map: 
            caller: 
        Returns:
        """
        return self.m_idx.get_local_map(_map,caller)

    def add_field(self, field):
        """         

          
           Just for good measure, to catch and short-circuit non-sensical uses
          
          virtual function void add_field  (uvm_reg_field field)
        Args:
            field: 
        """
        uvm_error(self.get_full_name(), "Cannot add field to an indirect data access register")


    #   virtual function void set (uvm_reg_data_t  value,
    #                              string          fname = "",
    #                              int             lineno = 0)
    #      `uvm_error(get_full_name(), "Cannot set() an indirect data access register")
    #   endfunction

    #
    #   virtual function uvm_reg_data_t  get(string  fname = "",
    #                                        int     lineno = 0)
    #      `uvm_error(get_full_name(), "Cannot get() an indirect data access register")
    #      return 0
    #   endfunction

    #
    #   virtual function uvm_reg get_indirect_reg(string  fname = "",
    #                                        int     lineno = 0)
    #      int unsigned idx = m_idx.get_mirrored_value()
    #      return(m_tbl[idx])
    #   endfunction

    def needs_update(self):
        return 0

    async def write(self, status, value, path=UVM_DEFAULT_PATH, _map=None,
            parent=None, prior=-1, extension=None, fname="", lineno=0):
        """         
          virtual task write(output uvm_status_e      status,
                             input  uvm_reg_data_t    value,
                             input  uvm_path_e        path = UVM_DEFAULT_PATH,
                             input  uvm_reg_map       map = None,
                             input  uvm_sequence_base parent = None,
                             input  int               prior = -1,
                             input  uvm_object        extension = None,
                             input  string            fname = "",
                             input  int               lineno = 0)
        Args:
            status: 
            value: 
            path: 
            UVM_DEFAULT_PATH: 
            _map: 
            parent: 
            prior: 
            extension: 
            fname: 
            lineno: 
        """
        uvm_check_output_args([status])

        if path == UVM_DEFAULT_PATH:
           blk = self.get_parent()  # uvm_reg_block
           path = blk.get_default_path()

        if path == UVM_BACKDOOR:
            uvm_warning(self.get_full_name(),
                "Cannot backdoor-write an indirect data access register. Switching to frontdoor.")
            path = UVM_FRONTDOOR


        # Can't simply call super.write() because it'll call set()
        # uvm_reg_item rw


        rw = UVMRegItem.type_id.create("write_item", None, self.get_full_name())
        await self.XatomicX(1, rw)
        rw.element      = self
        rw.element_kind = UVM_REG
        rw.kind         = UVM_WRITE
        rw.value[0]     = value
        rw.path         = path
        rw.map          = _map
        rw.parent       = parent
        rw.prior        = prior
        rw.extension    = extension
        rw.fname        = fname
        rw.lineno       = lineno

        await self.do_write(rw)
        status.append(rw.status)

        await self.XatomicX(0)
        #   endtask


    #   virtual task read(output uvm_status_e      status,
    #                     output uvm_reg_data_t    value,
    #                     input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                     input  uvm_reg_map       map = None,
    #                     input  uvm_sequence_base parent = None,
    #                     input  int               prior = -1,
    #                     input  uvm_object        extension = None,
    #                     input  string            fname = "",
    #                     input  int               lineno = 0)
    #
    #      if (path == UVM_DEFAULT_PATH):
    #         uvm_reg_block blk = get_parent()
    #         path = blk.get_default_path()
    #      end
    #
    #      if (path == UVM_BACKDOOR):
    #         `uvm_warning(get_full_name(), "Cannot backdoor-read an indirect data access register. Switching to frontdoor.")
    #         path = UVM_FRONTDOOR
    #      end
    #
    #      super.read(status, value, path, map, parent, prior, extension, fname, lineno)
    #   endtask


    #   virtual task poke(output uvm_status_e      status,
    #                     input  uvm_reg_data_t    value,
    #                     input  string            kind = "",
    #                     input  uvm_sequence_base parent = None,
    #                     input  uvm_object        extension = None,
    #                     input  string            fname = "",
    #                     input  int               lineno = 0)
    #      `uvm_error(get_full_name(), "Cannot poke() an indirect data access register")
    #      status = UVM_NOT_OK
    #   endtask

    #
    #   virtual task peek(output uvm_status_e      status,
    #                     output uvm_reg_data_t    value,
    #                     input  string            kind = "",
    #                     input  uvm_sequence_base parent = None,
    #                     input  uvm_object        extension = None,
    #                     input  string            fname = "",
    #                     input  int               lineno = 0)
    #      `uvm_error(get_full_name(), "Cannot peek() an indirect data access register")
    #      status = UVM_NOT_OK
    #   endtask

    #
    #   virtual task update(output uvm_status_e      status,
    #                       input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                       input  uvm_reg_map       map = None,
    #                       input  uvm_sequence_base parent = None,
    #                       input  int               prior = -1,
    #                       input  uvm_object        extension = None,
    #                       input  string            fname = "",
    #                       input  int               lineno = 0)
    #      status = UVM_IS_OK
    #   endtask

    #
    #   virtual task mirror(output uvm_status_e      status,
    #                       input uvm_check_e        check  = UVM_NO_CHECK,
    #                       input uvm_path_e         path = UVM_DEFAULT_PATH,
    #                       input uvm_reg_map        map = None,
    #                       input uvm_sequence_base  parent = None,
    #                       input int                prior = -1,
    #                       input  uvm_object        extension = None,
    #                       input string             fname = "",
    #                       input int                lineno = 0)
    #      status = UVM_IS_OK

    #   endtask
    #
    #endclass : uvm_reg_indirect_data


class uvm_reg_indirect_ftdr_seq(UVMRegFrontdoor):
    #   local uvm_reg m_addr_reg
    #   local uvm_reg m_data_reg
    #   local int     m_idx

    def __init__(self, addr_reg, idx, data_reg):
        """         
        Constructor
        Args:
            addr_reg: 
            idx: 
            data_reg: 
        """
        super().__init__("uvm_reg_indirect_ftdr_seq")
        self.m_addr_reg = addr_reg  # uvm_reg
        self.m_data_reg = data_reg  # uvm_reg
        self.m_idx      = idx       # int


    async def body(self):
        """             
        Body of indirect sequence
        """
        arr: List[UVMRegItem] = []
        if not sv.cast(arr, self.rw_info.clone(), UVMRegItem):
            uvm_fatal("CAST_FAIL", "Expected UVMRegItem, got " + str(self.rw_info))

        rw = arr[0]
        rw.element = self.m_addr_reg
        rw.kind    = UVM_WRITE
        rw.value[0] = self.m_idx

        await self.m_addr_reg.XatomicX(1, rw)
        await self.m_data_reg.XatomicX(1, rw)

        # This write selects the address to write/read
        await self.m_addr_reg.do_write(rw)

        if rw.status == UVM_NOT_OK:
            return

        arr = []
        if sv.cast(arr, self.rw_info.clone(), UVMRegItem):
            rw = arr[0]
        rw.element = self.m_data_reg

        # This fetches the actual data
        if self.rw_info.kind == UVM_WRITE:
            await self.m_data_reg.do_write(rw)
        else:
            await self.m_data_reg.do_read(rw)
            self.rw_info.value[0] = rw.value[0]

        await self.m_addr_reg.XatomicX(0)
        await self.m_data_reg.XatomicX(0)

        self.rw_info.status = rw.status




import unittest
from uvm.reg.uvm_reg import UVMReg
from uvm.reg.uvm_reg_block import UVMRegBlock
from uvm.reg.uvm_reg_model import *


class TestUVMRegBlock(unittest.TestCase):

    def test_add_reg(self):
        rb = UVMRegBlock("rb_blk")
        rb.create_map("", 0, 1, UVM_BIG_ENDIAN)
        reg1 = UVMReg("xx", 32, rb.get_full_name())
        reg1.configure(rb, None, "acp")
        # reg1.build()
        rb.default_map.add_reg(reg1, 0x0000, "RW")
        print("len is now " + str(len(rb.regs)))
        rr = rb.get_reg_by_name('xx')
        self.assertEqual(rr.get_name(), 'xx')

        r_none = rb.get_reg_by_name('y_reg')
        self.assertEqual(r_none, None)

    def test_add_block(self):
        top_b = UVMRegBlock("top_reg_block")
        top_b.create_map("custom_map", 0x100, 4, UVM_LITTLE_ENDIAN)
        custom_map = top_b.get_map_by_name("custom_map")
        top_b.set_default_map(custom_map)
        self.assertIsNotNone(custom_map)
        for i in range(10):
            base_addr = i * 0x100
            sub_b = UVMRegBlock("sub_blk_" + str(i))
            top_b.add_block(sub_b)
        top_b.configure(parent=None, hdl_path="")
        top_b.lock_model()


if __name__ == '__main__':
    unittest.main()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import RisingEdge, Timer, FallingEdge

from uvm.base import *
from uvm.comps import UVMDriver
from uvm.macros import uvm_component_utils

from ubus_transfer import *

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_slave_driver
#//
#//------------------------------------------------------------------------------
#class ubus_slave_driver extends uvm_driver #(ubus_transfer);
class ubus_slave_driver(UVMDriver):

    #  // new - constructor
    def __init__(self, name, parent):
        UVMDriver.__init__(self, name, parent)
        # The virtual interface used to drive and view HDL signals.
        self.vif = None
        self.tag = "UBUS_SLAVE_DRIVER_" + name

    def build_phase(self, phase):
        super().build_phase(phase)
        uvm_info(self.tag, "build_phase started", UVM_MEDIUM)
        arr = []
        if UVMConfigDb.get(self, "", "vif", arr):
            self.vif = arr[0]
        if self.vif is None:
            self.uvm_report_fatal("NOVIF", "virtual interface must be set for: " +
                self.get_full_name() + ".vif")

    # run phase
    
    async def run_phase(self, phase):
        #    fork
        fork_get = cocotb.fork(self.get_and_drive())
        fork_reset = cocotb.fork(self.reset_signals())
        #await [fork_get.join(), fork_reset.join()]
        await sv.fork_join([fork_get, fork_reset])
        #    join
    #  endtask : run_phase


    #  // get_and_drive
    #  virtual protected task get_and_drive();
    
    async def get_and_drive(self):
        await Timer(0)
        await FallingEdge(self.vif.sig_reset)
        while True:
            await RisingEdge(self.vif.sig_clock)
            req = []
            await self.seq_item_port.get_next_item(req)
            await self.respond_to_transfer(req[0])
            self.seq_item_port.item_done()
        #  endtask : get_and_drive


    #  // reset_signals
    
    async def reset_signals(self):
        while True:
            await RisingEdge(self.vif.sig_reset)
            self.vif.sig_error      <= 0
            self.vif.sig_wait       <= 0
            self.vif.slave_en       = 0


    #  // respond_to_transfer
    
    async def respond_to_transfer(self, resp):
        if resp.read_write != NOP:
            self.vif.sig_error <= 0
            for i in range(resp.size):
                if resp.read_write == READ:
                    self.vif.slave_en = 1
                    self.vif.sig_data_out <= resp.data[i]
                if resp.wait_state[i] > 0:
                    self.vif.sig_wait <= 1
                    for j in range(resp.wait_state[i]):
                        await RisingEdge(self.vif.sig_clock)
                self.vif.sig_wait <= 0
                await RisingEdge(self.vif.sig_clock)
                resp.data[i] = int(self.vif.sig_data)
            self.vif.slave_en = 0
            self.vif.sig_wait  <= 0  # 1'bz
            self.vif.sig_error <= 0  # 1'bz
        else:
            await Timer(0)
        #  endtask : respond_to_transfer

    #
    #endclass : ubus_slave_driver
#  Provide implementations of virtual methods such as get_type_name and create
uvm_component_utils(ubus_slave_driver)




import unittest
from uvm.seq.uvm_sequencer_base import UVMSequencerBase

from uvm.uvm_unit import MockObj

class MockSeq(MockObj):

    def __init__(self, name='mock_seq', parent=None):
        super().__init__(name, parent)
        self.parent = parent

    def get_parent_sequence(self):
        return self.parent


class TestUVMReg(unittest.TestCase):

    def test_init(self):
        sqr = UVMSequencerBase('new_sqr_123', None)
        self.assertFalse(sqr.is_grabbed())
        seq1 = MockSeq('seq1')
        seq2 = MockSeq('seq2')
        self.assertFalse(sqr.is_child(seq1, seq2))
        par_seq3 = MockSeq('par_seq3')
        seq4 = MockSeq('seq2', par_seq3)
        self.assertTrue(sqr.is_child(par_seq3, seq4))


if __name__ == '__main__':
    unittest.main()



#
#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2013 NVIDIA Corporation
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer
from .uvm_phase import UVMPhase, ph2str
from .uvm_object_globals import (UVM_DEBUG, UVM_PHASE_ENDED, UVM_PHASE_EXECUTING, UVM_PHASE_IMP,
                                 UVM_PHASE_READY_TO_END, UVM_PHASE_STARTED)
from .uvm_debug import uvm_debug
from .uvm_globals import uvm_zero_delay, uvm_report_fatal, uvm_report_info

#------------------------------------------------------------------------------
#
# Class: uvm_task_phase
#
#------------------------------------------------------------------------------
# Base class for all task phases.
# It forks a call to <uvm_phase::exec_task()>
# for each component in the hierarchy.
#
# The completion of the task does not imply, nor is it required for,
# the end of phase. Once the phase completes, any remaining forked
# <uvm_phase::exec_task()> threads are forcibly and immediately killed.
#
# By default, the way for a task phase to extend over time is if there is
# at least one component that raises an objection.
#| class my_comp extends uvm_component
#|    task main_phase(uvm_phase phase)
#|       phase.raise_objection(this, "Applying stimulus")
#|       ...
#|       phase.drop_objection(this, "Applied enough stimulus")
#|    endtask
#| endclass
#
#
# There is however one scenario wherein time advances within a task-based phase
# without any objections to the phase being raised. If two (or more) phases
# share a common successor, such as the <uvm_run_phase> and the
# <uvm_post_shutdown_phase> sharing the <uvm_extract_phase> as a successor,
# then phase advancement is delayed until all predecessors of the common
# successor are ready to proceed.  Because of this, it is possible for time to
# advance between <uvm_component::phase_started> and <uvm_component::phase_ended>
# of a task phase without any participants in the phase raising an objection.

class UVMTaskPhase(UVMPhase):

    def __init__(self, name):
        """         
          Function: new
         
          Create a new instance of a task-based phase
         
        Args:
            name: 
        """
        UVMPhase.__init__(self, name, UVM_PHASE_IMP)
        self.m_is_task_phase = True

    async def traverse(self, comp, phase, state):
        """         
          Function: traverse
         
          Traverses the component tree in bottom-up order, calling `execute` for
          each component. The actual order for task-based phases doesn't really
          matter, as each component task is executed in a separate process whose
          starting order is not deterministic.
        Args:
            comp: 
            phase: 
            state: 
        """
        phase.m_num_procs_not_yet_returned = 0
        await self.m_traverse(comp, phase, state)
        uvm_debug(self, 'traverse', 'Finished self.m_traverse for comp ' +
                comp.get_name())

    
    async def m_traverse(self, comp, phase, state):
        uvm_debug(self, "m_traverse", "START OF m_traverse, comp: " +
                comp.get_name())
        name = ""
        phase_domain = phase.get_domain()
        comp_domain = comp.get_domain()

        children = []
        comp.get_children(children)
        children = map(lambda c: c.get_name(), children)
        uvm_debug(self, "m_traverse", "Looping through children, comp: " +
                comp.get_name() + ' children: ' + ", ".join(children))

        # tpoikela: Added this loop, cause while-loop not safe
        children = []
        comp.get_children(children)
        for child in children:
            uvm_debug(self, "m_traverse", "Yielding now child traverse with "
                + child.get_name())
            await self.m_traverse(child, phase, state)

        # tpoikela: Not safe loop with parallel tasks phases
        #if comp.has_first_child():
        #    child = comp.get_first_child()
        #    while child is not None:
        #        uvm_debug(self, "m_traverse", "Yielding now child traverse with "
        #            + child.get_name())
        #        yield self.m_traverse(child, phase, state)
        #        child = comp.get_next_child()

        uvm_debug(self, "m_traverse", comp.get_name() + "| Comp children done.  Moving to its own phase..")

        if UVMPhase.m_phase_trace:
            dom_name = "unknown"
            if comp_domain is not None:
                dom_name = comp_domain.get_name()
            uvm_report_info("PH_TRACE",
                ("topdown-phase phase={} state={} comp={} comp.domain={} phase.domain={}".format(
                  phase.get_name(), str(state), comp.get_full_name(), dom_name,phase_domain.get_name()
                  )), UVM_DEBUG)

        from .uvm_domain import UVMDomain
        uvm_debug(self, 'm_traverse', "ph_dom: {}, comm_dom: {}".format(
            phase_domain.get_name(), UVMDomain.get_common_domain().get_name()))
        if (phase_domain == UVMDomain.get_common_domain() or phase_domain == comp_domain):
            uvm_debug(self, 'm_traverse', "Comp: " + comp.get_name() + " - " + self.get_name() +
                "| phase match found. Proceeding now...state is " + ph2str(state))
            if state == UVM_PHASE_STARTED:
                comp.m_current_phase = phase
                comp.m_apply_verbosity_settings(phase)
                comp.phase_started(phase)
                if hasattr(comp, 'm_sequencer_id'):
                    seqr = comp  # was if ($cast(seqr, comp))
                    uvm_debug(self, "m_traverse", comp.get_name() + " is SQR")
                    await seqr.start_phase_sequence(phase)
                else:
                    uvm_debug(self, "m_traverse", comp.get_name() + " is not SQR")
            elif state == UVM_PHASE_EXECUTING:
                ph = self  # uvm_phase
                if self in comp.m_phase_imps:
                    ph = comp.m_phase_imps[self]

                uvm_debug(self, "m_traverse", comp.get_name() + " yield ph.execute")
                await ph.execute(comp, phase)
            elif state == UVM_PHASE_READY_TO_END:
                comp.phase_ready_to_end(phase)
            elif state == UVM_PHASE_ENDED:
                uvm_debug(self, "m_traverse", "KKK")
                if hasattr(comp, 'm_sequencer_id'):
                    seqr = comp  # was if ($cast(seqr, comp))
                    seqr.stop_phase_sequence(phase)
                comp.phase_ended(phase)
                comp.m_current_phase = None
            else:
                uvm_report_fatal("PH_BADEXEC","task phase traverse internal error")
        uvm_debug(self, "m_traverse", "END OF m_traverse, comp: " +
                comp.get_name())

    async def execute(self, comp, phase):
        """         
          Function: execute
         
          Fork the task-based phase `phase` for the component `comp`.
         
        Args:
            comp: 
            phase: 
        """
        uvm_debug(self, 'execute', 'exec task_phase |' + self.get_name() + '| with comp: ' +
                comp.get_name())
        #fork
        #process proc
        # reseed this process for random stability
        #proc = process::self()
        #proc.srandom(uvm_create_random_seed(phase.get_type_name(), comp.get_full_name()))
        #phase.m_num_procs_not_yet_returned += 1
        proc = cocotb.fork(self._execute_fork_join_none(comp,phase))
        #phase.m_num_procs_not_yet_returned -= 1
        await uvm_zero_delay()
        #join_none


    async def _execute_fork_join_none(self, comp, phase):
        """             
        Args:
            comp: 
            phase: 
        """
        phase.m_num_procs_not_yet_returned += 1
        uvm_debug(self, '_execute_fork_join_none', 'exec task_phase |' + self.get_name()
                + '| yielding comp: ' + comp.get_name())
        await self.exec_task(comp, phase)
        uvm_debug(self, '_execute_fork_join_none', 'exec task_phase |' + self.get_name()
                + '| AFTER yield comp: ' + comp.get_name())
        phase.m_num_procs_not_yet_returned -= 1



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc. 
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from ..tlm1 import (UVMTLMFIFO, UVMAnalysisImp)
from ..macros import (uvm_fatal)


class UVMSequencerAnalysisFIFO(UVMTLMFIFO):

    #  uvm_analysis_imp #(RSP, UVMSequencerAnalysisFIFO #(RSP)) analysis_export
    #  uvm_sequencer_base sequencer_ptr


    def __init__(self, name, parent=None):
        super().__init__(name, parent, 0)
        self.analysis_export = UVMAnalysisImp("analysis_export", self)
        self.sequencer_ptr = None


    def write(self, t):
        if self.sequencer_ptr is None:
            uvm_fatal("SEQRNULL", "The sequencer pointer is None when attempting a write")
        self.sequencer_ptr.analysis_write(t)




import unittest
from uvm.base import (UVMComponent, UVMObjection)

class SubChild(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)

class Child(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.subchild = SubChild("subchild1", self)


class Parent(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.child = Child("child", self)
        self.child2 = Child("child2", self)


class TestUVMObjection(unittest.TestCase):


    def test_hier_obj(self):
        objection = UVMObjection("test_objection")
        parent = Parent("parent_objection_test", None)

        subchild1 = parent.child.subchild
        objection.raise_objection(subchild1, "subchild objection raised")
        self.assertEqual(objection.get_objection_count(subchild1), 1)
        self.assertEqual(objection.get_objection_total(parent.child), 1)
        self.assertEqual(objection.get_objection_total(parent), 1)

        objection.raise_objection(parent.child, "child objection raised")
        self.assertEqual(objection.get_objection_total(parent.child), 2)
        self.assertEqual(objection.get_objection_total(parent), 2)

        objection.raise_objection(parent.child2, "child2 objection raised")
        self.assertEqual(objection.get_objection_total(parent), 3)
        objection.raise_objection(parent, "parent objection raised")
        self.assertEqual(objection.get_objection_total(parent), 4)

        # Dropping the objections

        objection.drop_objection(subchild1, "subchild objection dropped")
        self.assertEqual(objection.get_objection_count(subchild1), 0)
        print(objection.convert2string())
        # self.assertEqual(objection.get_objection_total(parent.child), 1)
        # self.assertEqual(objection.get_objection_total(parent), 3)

        objection.drop_objection(parent, "parent objection dropped")
        self.assertEqual(objection.get_objection_count(parent), 0)

        objection.drop_objection(parent.child2, "child2 objection dropped")
        self.assertEqual(objection.get_objection_count(parent.child2), 0)

        objection.drop_objection(parent.child, "child objection dropped")
        self.assertEqual(objection.get_objection_count(parent.child), 0)
        # self.assertEqual(objection.get_objection_total(parent), 0)



#// 
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm import *

uvm_analysis_imp_sym_sb_expected = uvm_analysis_imp_decl("_sym_sb_expected")
uvm_analysis_imp_sym_sb_observed = uvm_analysis_imp_decl("_sym_sb_observed")

class sym_sb(UVMComponent):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.expected = uvm_analysis_imp_sym_sb_expected("expected", self)
        self.observed = uvm_analysis_imp_sym_sb_observed("observed", self)
        self.n_obs_thresh = 10
        self.m_sb = []
        self.error = 0
        self.m_n_obs = 0

    def build_phase(self, phase):
        thr = []
        if UVMConfigDb.get(self, "", "n_obs_thresh", thr):  # cast to 'void' removed
            self.n_obs_thresh = thr[0]


    def write_sym_sb_expected(self, tr):
        uvm_info("SB/EXP", sv.sformatf("Expected: 0x%h", tr.chr), UVM_MEDIUM)
        self.m_sb.append(tr.chr)


    def write_sym_sb_observed(self, tr):
        exp = 0x00
        uvm_info("SB/OBS", sv.sformatf("Observed: 0x%h", tr.chr), UVM_MEDIUM)
        exp = self.m_sb.pop(0)
        if tr.chr != exp:
            self.error = 1
            uvm_error("SB/MISMTCH", sv.sformatf("Symbol 0x%h observed instead of expected 0x%h",
                tr.chr, exp))
        self.m_n_obs += 1

    async def reset_phase(self, phase):
        self.m_n_obs = 0
        self.m_sb = []

    async def main_phase(self, phase):
        phase.raise_objection(self, "Have not checked enough data")
        while True:
            await Timer(250, "NS")
            if self.m_n_obs > self.n_obs_thresh:
                break
        phase.drop_objection(self, "Enough data has been observed")

    def check_phase(self, phase):
        if self.m_n_obs < self.n_obs_thresh:
            uvm_error("ERR/SCB", "Not enough items were observed")
            self.error = 1

uvm_component_utils(sym_sb)




#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

#
# About: 'ep' event poll
# This test is a simple test that will cover the creation of two uvm_events using the uvm_event_pool.
# Then use the uvm_event_pool methods to print those objects.
#
# For a full documentation of the uvm_event and uvm_event_pool, check the files:
#	- uvm/src/base/uvm_event.svh
#	- uvm/src/base/uvm_event.sv
#

import cocotb
from cocotb.triggers import Timer

from uvm.base.uvm_coreservice import UVMCoreService
from uvm.base.uvm_pool import UVMEventPool
from uvm.base.uvm_globals import UVM_FATAL, UVM_ERROR
from uvm.base.sv import sv

from uvm.macros import uvm_error
from uvm.base.uvm_global_vars import uvm_default_table_printer



async def trig_event(e):
    await Timer(100, "NS")
    e.trigger()
    await Timer(1, "NS")


async def trig_data_event(e):
    await Timer(50, "NS")
    e.trigger(0x1234)
    await Timer(1, "NS")


async def wait_on_evt(e):
    await e.wait_on()


@cocotb.test()
async def test_test(dut):

    ep = UVMEventPool("ep")
    cs_ = UVMCoreService.get()

    e = None
    e = ep.get("fred")
    e = ep.get("george")
    e_data = ep.get('evt_data')
    uvm_default_table_printer.knobs.reference = 0
    ep.print_obj()

    trig_proc = cocotb.fork(trig_event(e))
    cocotb.fork(trig_data_event(e_data))
    await e_data.wait_trigger()
    await e.wait_on()
    e.reset()
    await Timer(1, "NS")
    trig_proc = cocotb.fork(trig_event(e))
    evt_proc = cocotb.fork(wait_on_evt(e))
    await sv.fork_join([trig_proc, evt_proc])

    svr = None  # uvm_report_server
    svr = cs_.get_report_server()
    svr.report_summarize()
    time_ok = sv.realtime("NS") == 202
    data_ok = e_data.get_trigger_data() == 0x1234
    if data_ok and time_ok and (svr.get_severity_count(UVM_FATAL) +
            svr.get_severity_count(UVM_ERROR) == 0):
        print("** UVM TEST PASSED **\n")
    else:
        if data_ok is False:
            uvm_error("DATA_ERROR", "Exp: {}, Got: {}".format(0x1234,
                e_data.get_trigger_data()))
        if time_ok is False:
            uvm_error("TIME_ERROR", "Exp: {}, Got: {}".format(202,
                sv.realtime("NS")))
        raise Exception("!! UVM TEST FAILED !!\n")




import unittest
from uvm.base.uvm_component import UVMComponent
from uvm.base.uvm_topdown_phase import UVMTopdownPhase
from uvm.base.uvm_phase import UVMPhase
from uvm.base.uvm_object_globals import *


class TestUVMTopdownPhase(unittest.TestCase):


    def test_traverse(self):
        c1 = UVMComponent('uvm_test_top__test_uvm_topdown_phase', None)
        c2 = UVMComponent('sub_c1', c1)
        c3 = UVMComponent('sub_c2', c1)
        c4 = UVMComponent('sub_c2_c4', c2)
        td_phase = UVMTopdownPhase('MyPhase')
        states = [UVM_PHASE_STARTED, UVM_PHASE_READY_TO_END, UVM_PHASE_ENDED]
        phase = UVMPhase('MyActualPhase')
        for state in states:
            td_phase.traverse(c1, phase, state)

        children_c1 = []
        c1.get_children(children_c1)

        children_c2 = []
        c2.get_children(children_c2)
        self.assertEqual(len(children_c2), 1)
        self.assertEqual(children_c2[0].get_name(), c4.get_name())


if __name__ == '__main__':
    unittest.main()



#
#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------
"""
Analysis Ports

This section defines the port, export, and imp classes used for transaction
analysis.
"""

from ..macros.uvm_message_defines import uvm_fatal
from ..macros.uvm_tlm_defines import UVM_TLM_ANALYSIS_MASK
from ..base.uvm_port_base import UVM_UNBOUNDED_CONNECTIONS, UVMPortBase
from .uvm_tlm_imps import UVM_EXPORT, UVM_IMP_COMMON, UVM_PORT


class UVMAnalysisPort(UVMPortBase):
    """
    Broadcasts a value to all subscribers implementing a `UVMAnalysisImp`.

    .. code-block:: python
        class mon(UVMComponent):

            def __init__(self, name="sb", parent=None):
                super.new(name, parent)
                self.ap = UVMAnalysisPort("ap", self)

            async def run_phase(self, phase):
                t = None
                ...
                self.ap.write(t)
                ...
    """

    def __init__(self, name, parent):
        UVMPortBase.__init__(self, name, parent, UVM_PORT, 0, UVM_UNBOUNDED_CONNECTIONS)
        self.m_if_mask = UVM_TLM_ANALYSIS_MASK

    def get_type_name(self):
        return "uvm_analysis_port"

    def write(self, t):
        """
        Send specified value to all connected interface

        Args:
            t (any): Transaction to broadcast.
        """
        for i in range(0, self.size()):
            tif = self.get_if(i)
            if tif is None:
                uvm_fatal("NTCONN", ("No uvm_tlm interface is connected to " +
                    self.get_full_name() + " for executing write()"))
            else:
                tif.write(t)


class UVMAnalysisImp:
    """
    Receives all transactions broadcasted by a <uvm_analysis_port>. It serves as
    the termination point of an analysis port/export/imp connection. The component
    attached to the `imp` class--called a `subscriber`-- implements the analysis
    interface.

    Will invoke the ~write(T)~ method in the parent component.
    The implementation of the ~write(T)~ method must not modify
    the value passed to it.

    .. code-block:: python
      class sb(UVMComponent):
        uvm_analysis_imp#(trans, sb) ap

        def __init__(self, name="sb", parent=None):
           super().__init__(name, parent)
           self.ap = UVMAnalysisImp("ap", self)

        def write(self, t):
            ...

    """

    def write(self, t):
        """
        Args:
            t (any): Transaction to write.
        """
        self.m_imp.write(t)  # type: ignore


UVMAnalysisImp = UVM_IMP_COMMON(UVMAnalysisImp, UVM_TLM_ANALYSIS_MASK, "uvm_analysis_imp")


class UVMAnalysisExport(UVMPortBase):
    """
    Exports a lower-level `UVMAnalysisImp` to its parent. Export can be used to
    avoid long hierarchical paths like top.env.agent.monitor.
    """

    def __init__(self, name, parent=None):
        """
        Instantiate the export.

        Args:
            name (str): Name of the export.
            parent (UVMComponent): Parent component.
        """
        UVMPortBase.__init__(self, name, parent, UVM_EXPORT, 0, UVM_UNBOUNDED_CONNECTIONS)
        self.m_if_mask = UVM_TLM_ANALYSIS_MASK

    def get_type_name(self):
        return "uvm_analysis_export"

    def write(self, t):
        """
        Analysis port differs from other ports in that it broadcasts
        to all connected interfaces. Ports only send to the interface
        at the index specified in a call to set_if (0 by default).

        Args:
            t: Transaction to broadcast.
        """
        for i in range(0, self.size()):
            tif = self.get_if(i)
            if tif is None:
                uvm_fatal("NTCONN", ("No uvm_tlm interface is connected to "
                    + self.get_full_name() + " for executing write()"))
            else:
                tif.write(t)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//


import cocotb
from cocotb.triggers import FallingEdge
from cocotb.clock import Clock

from uvm import (UVMTest, uvm_top,
    UVMRegBitBashSeq, UVMMemWalkSeq,
    UVM_CVR_ALL, UVM_FINISHED, uvm_info, UVM_LOW,
    UVMReg, UVMCoreService, run_test, UVMConfigDb)

from tb_env import tb_env
from cocotb_coverage.coverage import coverage_db


class tb_test(UVMTest):

    def __init__(self, name="tb_test", parent=None):
        super().__init__(name, parent)
        self.dut = None

    def build_phase(self, phase):
        super().build_phase(phase)
        self.env = tb_env("env")

    
    async def run_phase(self, phase):
        env = self.env
        phase.raise_objection(self)

        env.regmodel.reset()
        env.regmodel.set_coverage(UVM_CVR_ALL)  # cast to 'void' removed

        self.dut.rst <= 0
        for _ in range(3):
            await FallingEdge(self.dut.clk)
        self.dut.rst <= 1
        await FallingEdge(self.dut.clk)

        seq = UVMRegBitBashSeq.type_id.create("reg_test_seq")
        seq.model = env.regmodel
        await seq.start(env.bus.sqr)
        await seq.wait_for_sequence_state(UVM_FINISHED)

        seq = UVMMemWalkSeq.type_id.create("memory_test_seq")
        seq.model = env.regmodel
        await seq.start(env.bus.sqr)
        await seq.wait_for_sequence_state(UVM_FINISHED)


        uvm_info("Test", "Generating and uploading 5 configurations...", UVM_LOW)

        for _ in range(5):
            status = 0
            #env.regmodel.randomize_with({Ra.F2.value == Rb.F2.value}) # TODO
            env.regmodel.randomize()
            status = []
            await env.regmodel.update(status)
            env.regmodel.sample_values()

        phase.drop_objection(self)


@cocotb.test()
async def initial_begin(dut):
    cs_ = UVMCoreService.get()

    UVMReg.include_coverage("*", UVM_CVR_ALL)
    c = Clock(dut.clk, 10, 'ns')

    test = tb_test("test")
    test.dut = dut

    svr = cs_.get_report_server()
    svr.set_max_quit_count(10)
    UVMConfigDb.set(None, "", "dut", dut)

    cocotb.fork(c.start())

    await run_test()

    def my_log(msg):
        uvm_info("COV_RPT", msg, UVM_LOW)

    coverage_db.report_coverage(my_log, bins=False)
    coverage_db.export_to_xml("results_coverage.xml")





import unittest

from uvm.reg.uvm_reg_field import UVMRegField, NO_RAND_SET
from uvm.reg.uvm_reg import UVMReg
from uvm.reg.uvm_reg_model import (UVM_READ, UVM_WRITE, UVM_PREDICT_WRITE,
    UVM_PREDICT_READ)


class TestUVMRegField(unittest.TestCase):

    def create_field(self, name, reg, acc='RW'):
        field = UVMRegField(name)
        field.configure(parent=reg, size=16, lsb_pos=0, access=acc,
            volatile=False, reset=0xABCD, has_reset=True, is_rand=False,
            individually_accessible=False)
        return field

    def test_configure(self):
        reg = UVMReg('my_reg', 16, False)
        field = self.create_field('f1', reg)
        self.assertEqual(field.get_reset(), 0xABCD)

    def test_string_conv(self):
        reg = UVMReg('r0_test', 16, False)
        field = self.create_field('f1', reg)
        ss = field.convert2string()
        self.assertRegex(ss, r'\[15:0\]={}'.format(16))

    def test_XupdateX(self):
        reg = UVMReg('update_reg', 16, False)
        field = self.create_field('field2', reg)
        field.set_access("RW")
        field.set(0x2345)
        value = field.XupdateX()
        self.assertEqual(value, 0x2345)
        # Clear on write
        field.set_access("WC")
        field.set(0x5555)
        value = field.XupdateX()
        self.assertEqual(value, 0x0)
        # Set on Write
        field.set_access("WS")
        field.set(0x5555)
        value = field.XupdateX()
        self.assertEqual(value, 0xFFFF)

    def test_set_get(self):
        for mode in NO_RAND_SET:
            reg = UVMReg('my_reg', 16, False)
            field = self.create_field('field_' + mode, reg, acc=mode)
            field.reset()
            access = field.get_access()
            field.set(0x1234)
            val = field.get()
            msg = 'Access ' + access + ' OK'
            predict_msg = 'Predict for ' + access + ' OK'
            if access in ['RW', 'WRS', 'WRC']:
                self.assertEqual(val, 0x1234, msg)
                self.assertTrue(field.predict(0x1234, -1, UVM_PREDICT_WRITE), predict_msg)
            elif access in ['RO', 'RC', 'RS']:
                self.assertEqual(val, field.get_reset(), msg)
                self.assertTrue(field.predict(field.get_reset(), -1, UVM_PREDICT_READ), predict_msg)
            else:
                self.assertTrue(field.predict(0x12), predict_msg)
                val = field.get()
                self.assertEqual(val, 0x12, 'Val OK after predict with ' + access)


if __name__ == '__main__':
    unittest.main()
